
# THIS FILE IS GENERATED FROM THEANO SETUP.PY
short_version = '0.6.0'
version = '0.6.0'
git_revision = '88f5a71d19a92305a82164c71f65ec8cdbf3a79c'
full_version = '0.6.0.dev-%(git_revision)s' % {
        'git_revision': git_revision}
release = True

if not release:
    version = full_version
